# helpers.py - helper functions for Flask-Restless
#
# Copyright 2011 Lincoln de Sousa <lincoln@comum.org>.
# Copyright 2012, 2013, 2014, 2015, 2016 Jeffrey Finkelstein
#           <jeffrey.finkelstein@gmail.com> and contributors.
#
# This file is part of Flask-Restless.
#
# Flask-Restless is distributed under both the GNU Affero General Public
# License version 3 and under the 3-clause BSD license. For more
# information, see LICENSE.AGPL and LICENSE.BSD.
"""Helper functions for Flask-Restless."""
import datetime
import inspect
from functools import lru_cache
from typing import List

from dateutil.parser import parse as parse_datetime
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import Interval
from sqlalchemy import Time
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.ext.hybrid import HYBRID_PROPERTY
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.inspection import inspect as sqlalchemy_inspect
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.orm import RelationshipProperty as RelProperty
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.attributes import QueryableAttribute
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import ColumnElement
from werkzeug.urls import url_quote_plus

try:
    # SQLAlchemy 1.3+
    from sqlalchemy.ext.associationproxy import ObjectAssociationProxyInstance as AssociationProxy
except ImportError:
    from sqlalchemy.ext.associationproxy import AssociationProxy

#: Names of attributes which should definitely not be considered relations when
#: dynamically computing a list of relations of a SQLAlchemy model.
RELATION_BLACKLIST = ('query', 'query_class', '_sa_class_manager',
                      '_decl_class_registry')

#: Types which should be considered columns of a model when iterating over all
#: attributes of a model class.
COLUMN_TYPES = (InstrumentedAttribute, hybrid_property)

#: Strings which, when received by the server as the value of a date or time
#: field, indicate that the server should use the current time when setting the
#: value of the field.
CURRENT_TIME_MARKERS = ('CURRENT_TIMESTAMP', 'CURRENT_DATE', 'LOCALTIMESTAMP')


def session_query(session, model):
    """Returns a SQLAlchemy query object for the specified `model`.

    If `model` has a ``query`` attribute already, ``model.query`` will be
    returned. If the ``query`` attribute is callable ``model.query()`` will be
    returned instead.

    If `model` has no such attribute, a query based on `session` will be
    created and returned.

    """
    if hasattr(model, 'query'):
        if callable(model.query):
            query = model.query()
        else:
            query = model.query
        if hasattr(query, 'filter'):
            return query
    return session.query(model)


def get_relations(model):
    """Returns a list of relation names of `model` (as a list of strings)."""
    return [k for k in dir(model)
            if not (k.startswith('_') or k in RELATION_BLACKLIST)
            and get_related_model(model, k)]


@lru_cache()
def get_related_model(model, relationname):
    """Gets the class of the model to which `model` is related by the attribute
    whose name is `relationname`.

    For example, if we have the model classes ::

        class Person(Base):
            __tablename__ = 'person'
            id = Column(Integer, primary_key=True)
            articles = relationship('Article')

        class Article(Base):
            __tablename__ = 'article'
            id = Column(Integer, primary_key=True)
            author_id = Column(Integer, ForeignKey('person.id'))
            author = relationship('Person')

    then

        >>> get_related_model(Person, 'articles')
        <class 'Article'>
        >>> get_related_model(Article, 'author')
        <class 'Person'>

    """
    if hasattr(model, relationname):
        # inspector = sqlalchemy_inspect(model)
        # attributes = inspector.attrs
        # if relationname in attributes:
        #     state = attributes[relationname]
        attr = getattr(model, relationname)
        if hasattr(attr, 'property') \
                and isinstance(attr.property, RelProperty):
            return attr.property.mapper.class_
        if isinstance(attr, AssociationProxy):
            return get_related_association_proxy_model(attr)
    return None


def get_related_association_proxy_model(attr):
    """Returns the model class specified by the given SQLAlchemy relation
    attribute, or ``None`` if no such class can be inferred.

    `attr` must be a relation attribute corresponding to an association proxy.

    """
    prop = attr.remote_attr.property
    for attribute in ('mapper', 'parent'):
        if hasattr(prop, attribute):
            return getattr(prop, attribute).class_
    return None


def foreign_key_columns(model):
    """Returns a list of the :class:`sqlalchemy.Column` objects that contain
    foreign keys for relationships in the specified model class.

    """
    try:
        inspector = sqlalchemy_inspect(model)
    except NoInspectionAvailable:
        # Well, the inspection of a model class returns a mapper anyway, so
        # let's just assume the inspection would have returned the mapper.
        inspector = class_mapper(model)
    all_columns = inspector.columns
    return [c for c in all_columns if c.foreign_keys]


def foreign_keys(model):
    """Returns a list of the names of columns that contain foreign keys for
    relationships in the specified model class.

    """
    return [column.name for column in foreign_key_columns(model)]


def has_field(model, fieldname):
    """Returns ``True`` if the `model` has the specified field or if it has a
    settable hybrid property for this field name.

    """
    descriptors = sqlalchemy_inspect(model).all_orm_descriptors._data
    if fieldname in descriptors and hasattr(descriptors[fieldname], 'fset'):
        return descriptors[fieldname].fset is not None
    return hasattr(model, fieldname)


def get_field_type(model, fieldname):
    """Helper which returns the SQLAlchemy type of the field."""
    field = getattr(model, fieldname)
    if isinstance(field, ColumnElement):
        return field.type
    if isinstance(field, AssociationProxy):
        field = field.remote_attr
    if hasattr(field, 'property'):
        prop = field.property
        if isinstance(prop, RelProperty):
            return None
        return prop.columns[0].type
    return None


def attribute_columns(model) -> List[str]:
    """Returns a list of model's column names that should be considered as attributes."""
    inspected_model = sqlalchemy_inspect(model)
    column_attrs = inspected_model.column_attrs.keys()
    descriptors = inspected_model.all_orm_descriptors.items()
    hybrid_columns = [k for k, d in descriptors if d.extension_type == HYBRID_PROPERTY]

    return column_attrs + hybrid_columns


@lru_cache()
def primary_key_names(model):
    """Returns all the primary keys for a model."""
    return [key for key, field in inspect.getmembers(model)
            if isinstance(field, QueryableAttribute)
            and hasattr(field, 'property')
            and isinstance(field.property, ColumnProperty)
            and field.property.columns[0].primary_key]


def primary_key_value(instance, as_string=False):
    """Returns the value of the primary key field of the specified `instance`
    of a SQLAlchemy model.

    This is a convenience function for::

        getattr(instance, primary_key_name(instance))

    If `as_string` is ``True``, try to coerce the return value to a string.

    """
    result = getattr(instance, primary_key_for(instance))
    if not as_string:
        return result
    try:
        return str(result)
    except UnicodeEncodeError:
        return url_quote_plus(result.encode('utf-8'))


def is_like_list(instance, relation):
    """Returns ``True`` if and only if the relation of `instance` whose name is
    `relation` is list-like.

    A relation may be like a list if, for example, it is a non-lazy one-to-many
    relation, or it is a dynamically loaded one-to-many.

    """
    if relation in instance._sa_class_manager:
        return instance._sa_class_manager[relation].property.uselist
    elif hasattr(instance, relation):
        attr = getattr(instance._sa_instance_state.class_, relation)
        if hasattr(attr, 'property'):
            return attr.property.uselist
    related_value = getattr(type(instance), relation, None)
    if isinstance(related_value, AssociationProxy):
        local_prop = related_value.local_attr.prop
        if isinstance(local_prop, RelProperty):
            return local_prop.uselist
    return False


def is_mapped_class(cls):
    """Returns ``True`` if and only if the specified SQLAlchemy model class is
    a mapped class.

    """
    try:
        sqlalchemy_inspect(cls)
    except NoInspectionAvailable:
        return False
    else:
        return True


def query_by_primary_key(session, model, pk_value, primary_key=None):
    """Returns a SQLAlchemy query object containing the result of querying
    `model` for instances whose primary key has the value `pk_value`.

    If `primary_key` is specified, the column specified by that string is used
    as the primary key column. Otherwise, the column named ``id`` is used.

    Presumably, the returned query should have at most one element.

    """
    pk_name = primary_key or primary_key_for(model)
    query = session_query(session, model)
    return query.filter(getattr(model, pk_name) == pk_value)


def get_by(session, model, pk_value, primary_key=None):
    """Returns the first instance of `model` whose primary key has the value
    `pk_value`, or ``None`` if no such instance exists.

    If `primary_key` is specified, the column specified by that string is used
    as the primary key column. Otherwise, the column named ``id`` is used.

    """
    result = query_by_primary_key(session, model, pk_value, primary_key)
    return result.first()


def string_to_datetime(model, fieldname, value):
    """Casts `value` to a :class:`datetime.datetime` or
    :class:`datetime.timedelta` object if the given field of the given
    model is a date-like or interval column.

    If the field name corresponds to a field in the model which is a
    :class:`sqlalchemy.types.Date`, :class:`sqlalchemy.types.DateTime`,
    or :class:`sqlalchemy.Interval`, then the returned value will be the
    :class:`datetime.datetime` or :class:`datetime.timedelta` Python
    object corresponding to `value`. Otherwise, the `value` is returned
    unchanged.

    """
    if value is None:
        return value
    # If this is a date, time or datetime field, parse it and convert it to
    # the appropriate type.
    field_type = get_field_type(model, fieldname)
    if isinstance(field_type, (Date, Time, DateTime)):
        # If the string is empty, no datetime can be inferred from it.
        if value.strip() == '':
            return None
        # If the string is a string indicating that the value of should be the
        # current datetime on the server, get the current datetime that way.
        if value in CURRENT_TIME_MARKERS:
            return getattr(func, value.lower())()
        value_as_datetime = parse_datetime(value)
        # If the attribute on the model needs to be a Date or Time object as
        # opposed to a DateTime object, just get the date component of the
        # datetime.
        if isinstance(field_type, Date):
            return value_as_datetime.date()
        if isinstance(field_type, Time):
            return value_as_datetime.timetz()
        return value_as_datetime
    # If this is an Interval field, convert the integer value to a timedelta.
    if isinstance(field_type, Interval) and isinstance(value, int):
        return datetime.timedelta(seconds=value)
    # In any other case, simply copy the value unchanged.
    return value


def strings_to_datetimes(model, dictionary):
    """Returns a new dictionary with all the mappings of `dictionary` but
    with date strings and intervals mapped to :class:`datetime.datetime` or
    :class:`datetime.timedelta` objects.

    The keys of `dictionary` are names of fields in the model specified in the
    constructor of this class. The values are values to set on these fields. If
    a field name corresponds to a field in the model which is a
    :class:`sqlalchemy.types.Date`, :class:`sqlalchemy.types.DateTime`, or
    :class:`sqlalchemy.Interval`, then the returned dictionary will have the
    corresponding :class:`datetime.datetime` or :class:`datetime.timedelta`
    Python object as the value of that mapping in place of the string.

    This function outputs a new dictionary; it does not modify the argument.

    """
    return {k: string_to_datetime(model, k, v) for k, v in dictionary.items() if k not in ('type', 'links')}


def get_model(instance):
    """Returns the model class of which the specified object is an instance."""
    return type(instance)


# This code comes from <http://stackoverflow.com/a/6798042/108197>, which is
# licensed under the Creative Commons Attribution-ShareAlike License version
# 3.0 Unported.
#
# That is an answer originally authored by the user
# <http://stackoverflow.com/users/500584/agf> to the question
# <http://stackoverflow.com/q/6760685/108197>.
#
# TODO This code is for simultaneous Python 2 and 3 usage. It can be greatly
# simplified when removing Python 2 support.
class _Singleton(type):
    """A metaclass for a singleton class."""

    #: The known instances of the class instantiating this metaclass.
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Returns the singleton instance of the specified class."""
        if cls not in cls._instances:
            supercls = super(_Singleton, cls)
            cls._instances[cls] = supercls.__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(_Singleton('SingletonMeta', (object,), {})):
    """Base class for a singleton class."""
    pass


class KnowsAPIManagers:
    """An object that allows client code to register :class:`APIManager`
    objects.

    """

    def __init__(self):
        #: A global list of created :class:`APIManager` objects.
        self.created_managers = set()

    def register(self, apimanager):
        """Inform this object about the specified :class:`APIManager` object.

        """
        self.created_managers.add(apimanager)


class ModelFinder(KnowsAPIManagers, Singleton):
    """The singleton class that backs the :func:`model_for` function."""

    @lru_cache()
    def __call__(self, resource_type, _apimanager=None, **kw):
        if _apimanager is not None:
            # This may raise ValueError.
            return _apimanager.model_for(resource_type, **kw)
        for manager in self.created_managers:
            try:
                return self(resource_type, _apimanager=manager, **kw)
            except ValueError:
                pass
        message = ('No model with collection name {0} is known to any'
                   ' APIManager objects; maybe you have not set the'
                   ' `collection_name` keyword argument when calling'
                   ' `APIManager.create_api()`?').format(resource_type)
        raise ValueError(message)


class CollectionNameFinder(KnowsAPIManagers, Singleton):
    """The singleton class that backs the :func:`collection_name` function."""

    @lru_cache()
    def __call__(self, model, _apimanager=None, **kw):
        if _apimanager is not None:
            if model not in _apimanager.created_apis_for:
                message = ('APIManager {0} has not created an API for model '
                           ' {1}').format(_apimanager, model)
                raise ValueError(message)
            return _apimanager.collection_name(model, **kw)
        for manager in self.created_managers:
            try:
                return self(model, _apimanager=manager, **kw)
            except ValueError:
                pass
        message = ('Model {0} is not known to any APIManager'
                   ' objects; maybe you have not called'
                   ' APIManager.create_api() for this model.').format(model)
        raise ValueError(message)


class UrlFinder(KnowsAPIManagers, Singleton):
    """The singleton class that backs the :func:`url_for` function."""

    def __call__(self, model, resource_id=None, relation_name=None,
                 related_resource_id=None, _apimanager=None,
                 relationship=False, **kw):
        if _apimanager is not None:
            if model not in _apimanager.created_apis_for:
                message = ('APIManager {0} has not created an API for model '
                           ' {1}; maybe another APIManager instance'
                           ' did?').format(_apimanager, model)
                raise ValueError(message)
            return _apimanager.url_for(model, resource_id=resource_id,
                                       relation_name=relation_name,
                                       related_resource_id=related_resource_id,
                                       relationship=relationship, **kw)
        for manager in self.created_managers:
            try:
                return self(model, resource_id=resource_id,
                            relation_name=relation_name,
                            related_resource_id=related_resource_id,
                            relationship=relationship, _apimanager=manager,
                            **kw)
            except ValueError:
                pass
        message = ('Model {0} is not known to any APIManager'
                   ' objects; maybe you have not called'
                   ' APIManager.create_api() for this model.').format(model)
        raise ValueError(message)


class SerializerFinder(KnowsAPIManagers, Singleton):
    """The singleton class that backs the :func:`serializer_for` function."""

    @lru_cache()
    def __call__(self, model, _apimanager=None, **kw):
        if _apimanager is not None:
            if model not in _apimanager.created_apis_for:
                message = ('APIManager {0} has not created an API for model '
                           ' {1}').format(_apimanager, model)
                raise ValueError(message)
            return _apimanager.serializer_for(model, **kw)
        for manager in self.created_managers:
            try:
                return self(model, _apimanager=manager, **kw)
            except ValueError:
                pass
        message = ('Model {0} is not known to any APIManager'
                   ' objects; maybe you have not called'
                   ' APIManager.create_api() for this model.').format(model)
        raise ValueError(message)


class PrimaryKeyFinder(KnowsAPIManagers, Singleton):
    """The singleton class that backs the :func:`primary_key_for` function."""

    @lru_cache()
    def __call__(self, instance_or_model, _apimanager=None, **kw):
        if isinstance(instance_or_model, type):
            model = instance_or_model
        else:
            model = instance_or_model.__class__

        if _apimanager is not None:
            managers_to_search = [_apimanager]
        else:
            managers_to_search = self.created_managers
        for manager in managers_to_search:
            if model in manager.created_apis_for:
                primary_key = manager.primary_key_for(model, **kw)
                break
        else:
            message = ('Model "{0}" is not known to {1}; maybe you have not'
                       ' called APIManager.create_api() for this model?')
            if _apimanager is not None:
                manager_string = 'APIManager "{0}"'.format(_apimanager)
            else:
                manager_string = 'any APIManager objects'
            message = message.format(model, manager_string)
            raise ValueError(message)

        # If `APIManager.create_api(model)` was called without providing
        # a value for the `primary_key` keyword argument, then we must
        # compute the primary key name from the model directly.
        if primary_key is None:
            pk_names = primary_key_names(model)
            primary_key = 'id' if 'id' in pk_names else pk_names[0]
        return primary_key


#: Returns the URL for the specified model, similar to :func:`flask.url_for`.
#:
#: `model` is a SQLAlchemy model class. This should be a model on which
#: :meth:`APIManager.create_api_blueprint` (or :meth:`APIManager.create_api`)
#: has been invoked previously. If no API has been created for it, this
#: function raises a `ValueError`.
#:
#: If `_apimanager` is not ``None``, it must be an instance of
#: :class:`APIManager`. Restrict our search for endpoints exposing `model` to
#: only endpoints created by the specified :class:`APIManager` instance.
#:
#: The `resource_id`, `relation_name`, and `relationresource_id` keyword
#: arguments allow you to get the URL for a more specific sub-resource.
#:
#: For example, suppose you have a model class ``Person`` and have created the
#: appropriate Flask application and SQLAlchemy session::
#:
#:     >>> manager = APIManager(app, session=session)
#:     >>> manager.create_api(Person, collection_name='people')
#:     >>> url_for(Person, resource_id=3)
#:     'http://example.com/api/people/3'
#:     >>> url_for(Person, resource_id=3, relation_name=computers)
#:     'http://example.com/api/people/3/computers'
#:     >>> url_for(Person, resource_id=3, relation_name=computers, related_resource_id=9)
#:     'http://example.com/api/people/3/computers/9'
#:
#: If a `resource_id` and a `relation_name` are provided, and you wish
#: to determine the relationship endpoint URL instead of the related
#: resource URL, set the `relationship` keyword argument to ``True``::
#:
#:     >>> url_for(Person, resource_id=3, relation_name=computers, relationshi=True)
#:     'http://example.com/api/people/3/relatonships/computers'
#:
#: The remaining keyword arguments, `kw`, are passed directly on to
#: :func:`flask.url_for`.
#:
#: Since this function creates absolute URLs to resources linked to the given
#: instance, it must be called within a `Flask request context`_.
#:
#:  .. _Flask request context: http://flask.pocoo.org/docs/0.10/reqcontext/
#:
url_for = UrlFinder()

#: Returns the collection name for the specified model, as specified by the
#: ``collection_name`` keyword argument to :meth:`APIManager.create_api` when
#: it was previously invoked on the model.
#:
#: `model` is a SQLAlchemy model class. This should be a model on which
#: :meth:`APIManager.create_api_blueprint` (or :meth:`APIManager.create_api`)
#: has been invoked previously. If no API has been created for it, this
#: function raises a `ValueError`.
#:
#: If `_apimanager` is not ``None``, it must be an instance of
#: :class:`APIManager`. Restrict our search for endpoints exposing `model` to
#: only endpoints created by the specified :class:`APIManager` instance.
#:
#: For example, suppose you have a model class ``Person`` and have created the
#: appropriate Flask application and SQLAlchemy session::
#:
#:     >>> from mymodels import Person
#:     >>> manager = APIManager(app, session=session)
#:     >>> manager.create_api(Person, collection_name='people')
#:     >>> collection_name(Person)
#:     'people'
#:
#: This function is the inverse of :func:`model_for`::
#:
#:     >>> manager.collection_name(manager.model_for('people'))
#:     'people'
#:     >>> manager.model_for(manager.collection_name(Person))
#:     <class 'mymodels.Person'>
#:
collection_name = CollectionNameFinder()

#: Returns the callable serializer object for the specified model, as
#: specified by the `serializer` keyword argument to
#: :meth:`APIManager.create_api` when it was previously invoked on the
#: model.
#:
#: `model` is a SQLAlchemy model class. This should be a model on which
#: :meth:`APIManager.create_api_blueprint` (or :meth:`APIManager.create_api`)
#: has been invoked previously. If no API has been created for it, this
#: function raises a `ValueError`.
#:
#: If `_apimanager` is not ``None``, it must be an instance of
#: :class:`APIManager`. Restrict our search for endpoints exposing
#: `model` to only endpoints created by the specified
#: :class:`APIManager` instance.
#:
#: For example, suppose you have a model class ``Person`` and have
#: created the appropriate Flask application and SQLAlchemy session::
#:
#:     >>> from mymodels import Person
#:     >>> def my_serializer(model, *args, **kw):
#:     ...     # return something cool here...
#:     ...     return {}
#:     ...
#:     >>> manager = APIManager(app, session=session)
#:     >>> manager.create_api(Person, serializer=my_serializer)
#:     >>> serializer_for(Person)
#:     <function my_serializer at 0x...>
#:
serializer_for = SerializerFinder()

#: Returns the model corresponding to the given collection name, as specified
#: by the ``collection_name`` keyword argument to :meth:`APIManager.create_api`
#: when it was previously invoked on the model.
#:
#: `collection_name` is a string corresponding to the "type" of a model. This
#: should be a model on which :meth:`APIManager.create_api_blueprint` (or
#: :meth:`APIManager.create_api`) has been invoked previously. If no API has
#: been created for it, this function raises a `ValueError`.
#:
#: If `_apimanager` is not ``None``, it must be an instance of
#: :class:`APIManager`. Restrict our search for endpoints exposing `model` to
#: only endpoints created by the specified :class:`APIManager` instance.
#:
#: For example, suppose you have a model class ``Person`` and have created the
#: appropriate Flask application and SQLAlchemy session::
#:
#:     >>> from mymodels import Person
#:     >>> manager = APIManager(app, session=session)
#:     >>> manager.create_api(Person, collection_name='people')
#:     >>> model_for('people')
#:     <class 'mymodels.Person'>
#:
#: This function is the inverse of :func:`collection_name`::
#:
#:     >>> manager.collection_name(manager.model_for('people'))
#:     'people'
#:     >>> manager.model_for(manager.collection_name(Person))
#:     <class 'mymodels.Person'>
#:
model_for = ModelFinder()

#: Returns the primary key to be used for the given model or model instance,
#: as specified by the ``primary_key`` keyword argument to
#: :meth:`APIManager.create_api` when it was previously invoked on the model.
#:
#: `primary_key` is a string corresponding to the primary key identifier
#: to be used by flask-restless for a model. If no primary key has been set
#: at the flask-restless level (by using the ``primary_key`` keyword argument
#: when calling :meth:`APIManager.create_api_blueprint`, the model's primary
#: key will be returned. If no API has been created for the model, this
#: function raises a `ValueError`.
#:
#: If `_apimanager` is not ``None``, it must be an instance of
#: :class:`APIManager`. Restrict our search for endpoints exposing `model` to
#: only endpoints created by the specified :class:`APIManager` instance.
#:
#: For example, suppose you have a model class ``Person`` and have created the
#: appropriate Flask application and SQLAlchemy session::
#:
#:     >>> from mymodels import Person
#:     >>> manager = APIManager(app, session=session)
#:     >>> manager.create_api(Person, primary_key='name')
#:     >>> primary_key_for(Person)
#:     'name'
#:     >>> my_person = Person(name="Bob")
#:     >>> primary_key_for(my_person)
#:     'name'
#:
#: This is in contrast to the typical default:
#:
#:     >>> manager = APIManager(app, session=session)
#:     >>> manager.create_api(Person)
#:     >>> primary_key_for(Person)
#:     'id'
#:
primary_key_for = PrimaryKeyFinder()



# test_filtering.py - unit tests for filtering resources in client requests
#
# Copyright 2011 Lincoln de Sousa <lincoln@comum.org>.
# Copyright 2012, 2013, 2014, 2015, 2016 Jeffrey Finkelstein
#           <jeffrey.finkelstein@gmail.com> and contributors.
#
# This file is part of Flask-Restless.
#
# Flask-Restless is distributed under both the GNU Affero General Public
# License version 3 and under the 3-clause BSD license. For more
# information, see LICENSE.AGPL and LICENSE.BSD.
"""Unit tests for filtering resources in client requests."""

import unittest
from datetime import date
from datetime import datetime
from datetime import time

from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Time
from sqlalchemy import Unicode
from sqlalchemy.dialects.postgresql import INET
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from testing.postgresql import PostgresqlFactory as PGFactory

from .helpers import ManagerTestBase
from .helpers import check_sole_error
from .helpers import dumps
from .helpers import loads

# This import is unused but is required for testing on PyPy. CPython can
# use psycopg2, but PyPy can only use psycopg2cffi.
try:
    import psycopg2  # noqa
except ImportError:
    from psycopg2cffi import compat
    compat.register()



#: The PostgreSQL class used to create a temporary database for testing.
#:
#: This class should be instantiated in the setUp method of test
#: classes, and the :class:`Postgresql.stop` method should be called on
#: tearDown.
#:
#: This is an optimization designed to speed up the tests that require
#: PostgreSQL, since it can be extremely slow to initialize a PostgreSQL
#: database before each test method.
PostgreSQL = PGFactory(cache_initialized_db=True)


def tearDown():
    """Clears the cache in the :attr:`PostgreSQL` class."""
    PostgreSQL.clear_cache()


class SearchTestBase(ManagerTestBase):
    """Provides a search method that simplifies a fetch request with filtering
    query parameters.

    """

    def search(self, url, filters=None, single=None):
        """Convenience function for performing a filtered :http:method:`get`
        request.

        `url` is the ``path`` part of the URL to which the request will be
        sent.

        If `filters` is specified, it must be a Python list containing filter
        objects. It specifies how to set the ``filter[objects]`` query
        parameter.

        If `single` is specified, it must be a Boolean. It specifies how to set
        the ``filter[single]`` query parameter.

        """
        if filters is None:
            filters = []
        params = {'filter[objects]': dumps(filters)}
        if single is not None:
            params['filter[single]'] = 1 if single else 0
        return self.app.get(url, query_string=params)


class TestFiltering(SearchTestBase):
    """Tests for filtering resources.

    For more information, see the `Filtering`_ section of the JSON API
    specification.

    .. _Filtering: http://jsonapi.org/format/#fetching-filtering

    """

    def setUp(self):
        """Creates the database, the :class:`~flask.Flask` object, the
        :class:`~flask_restless.manager.APIManager` for that application,
        and creates the ReSTful API endpoints for the models used in the test
        methods.

        """
        super(TestFiltering, self).setUp()

        class Article(self.Base):
            __tablename__ = 'article'
            id = Column(Integer, primary_key=True)
            author_id = Column(Integer, ForeignKey('person.id'))
            author = relationship('Person', backref=backref('articles'))

        class Person(self.Base):
            __tablename__ = 'person'
            id = Column(Integer, primary_key=True)
            name = Column(Unicode)
            age = Column(Integer)
            birthday = Column(Date)
            birth_datetime = Column(DateTime)
            bedtime = Column(Time)

        class Comment(self.Base):
            __tablename__ = 'comment'
            id = Column(Integer, primary_key=True)
            content = Column(Unicode)
            article_id = Column(Integer, ForeignKey('article.id'))
            article = relationship(Article, backref=backref('comments'))
            author_id = Column(Integer, ForeignKey('person.id'))
            author = relationship(Person, backref=backref('comments'))

        self.Article = Article
        self.Person = Person
        self.Comment = Comment
        self.Base.metadata.create_all()
        self.manager.create_api(Article)
        self.manager.create_api(Person)
        # HACK Need to create APIs for these other models because otherwise
        # we're not able to create the link URLs to them.
        #
        # TODO Fix this by simply not creating links to related models for
        # which no API has been made.
        self.manager.create_api(Comment)

    def test_bad_filter(self):
        """Tests that providing a bad filter parameter causes an error
        response.

        """
        query_string = {'filter[objects]': 'bogus'}
        response = self.app.get('/api/person', query_string=query_string)
        assert response.status_code == 400
        # TODO check error messages here

    def test_bad_filter_relation(self):
        """Tests for providing a bad filter parameter for fetching a
        relation.

        """
        query_string = {'filter[objects]': 'bogus'}
        response = self.app.get('/api/person/1/articles',
                                query_string=query_string)
        check_sole_error(response, 400, ['Unable to decode', 'filter object'])

    def test_bad_filter_relationship(self):
        """Test for providing a bad filter parameter for fetching
        relationship objects.

        """
        person = self.Person(id=1)
        self.session.add(person)
        self.session.commit()
        query_string = {'filter[objects]': 'bogus'}
        response = self.app.get('/api/person/1/relationships/articles',
                                query_string=query_string)
        assert response.status_code == 400
        # TODO check error messages here

    def test_like(self):
        """Tests for filtering using the ``like`` operator."""
        person1 = self.Person(name=u'Jesus')
        person2 = self.Person(name=u'Mary')
        person3 = self.Person(name=u'Joseph')
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='name', op='like', val='%s%')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert len(people) == 2
        assert ['Jesus', 'Joseph'] == sorted(person['attributes']['name']
                                             for person in people)

    def test_single(self):
        """Tests for requiring a single resource response to a filtered
        request.

        """
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='id', op='equals', val='1')]
        response = self.search('/api/person', filters, single=True)
        assert response.status_code == 200
        document = loads(response.data)
        person = document['data']
        assert person['id'] == '1'

    def test_single_relationship(self):
        """Tests for requiring a single relationship object in a
        response to a filtered request.

        """
        person = self.Person(id=1)
        article1 = self.Article(id=1)
        article2 = self.Article(id=2)
        person.articles = [article1, article2]
        self.session.add_all([person, article1, article2])
        self.session.commit()
        filters = [dict(name='id', op='equals', val='1')]
        response = self.search('/api/person/1/relationships/articles', filters,
                               single=True)
        assert response.status_code == 200
        document = loads(response.data)
        article = document['data']
        # Check that this is just a resource identifier object and not a
        # full resource object representing the Article.
        assert ['id', 'type'] == sorted(article)
        assert article['type'] == 'article'
        assert article['id'] == '1'

    def test_single_too_many(self):
        """Tests that requiring a single resource response returns an error if
        the filtered request would have returned more than one resource.

        """
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        self.session.add_all([person1, person2])
        self.session.commit()
        response = self.search('/api/person', single=True)
        # TODO should this be a 404? Maybe 409 is better?
        assert response.status_code == 404
        # TODO check the error message here.

    def test_single_too_few(self):
        """Tests that requiring a single resource response yields an error
        response if the filtered request would have returned zero resources.

        """
        response = self.search('/api/person', single=True)
        # TODO should this be a 404? Maybe 409 is better?
        assert response.status_code == 404
        # TODO check the error message here.

    def test_single_wrong_format(self):
        """Tests that providing an incorrectly formatted argument to
        ``filter[single]`` yields an error response.

        """
        params = {'filter[single]': 'bogus'}
        response = self.app.get('/api/person', query_string=params)
        assert response.status_code == 400
        # TODO check the error message here.

    def test_relation_single_wrong_format(self):
        """Tests that providing an incorrectly formatted argument to
        ``filter[single]`` yields an error response when fetching a
        relation.

        """
        person = self.Person(id=1)
        self.session.add(person)
        self.session.commit()
        params = {'filter[single]': 'bogus'}
        response = self.app.get('/api/person/1/articles', query_string=params)
        assert response.status_code == 400
        # TODO check the error message here.

    def test_relationship_single_wrong_format(self):
        """Tests that providing an incorrectly formatted argument to
        ``filter[single]`` yields an error response when fetching
        relationship.

        """
        person = self.Person(id=1)
        self.session.add(person)
        self.session.commit()
        query_string = {'filter[single]': 'bogus'}
        response = self.app.get('/api/person/1/relationships/articles',
                                query_string=query_string)
        assert response.status_code == 400
        # TODO check the error message here.

    def test_in_list(self):
        """Tests for a filter object checking for a field with value in a
        specified list of acceptable values.

        """
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='id', op='in', val=[2, 3])]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert len(people) == 2
        assert ['2', '3'] == sorted(person['id'] for person in people)

    def test_any_in_to_many(self):
        """Test for filtering using the ``any`` operator with a sub-filter
        object on a to-many relationship.

        """
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        comment1 = self.Comment(content=u"that's cool!", author=person1)
        comment2 = self.Comment(content=u'i like turtles', author=person2)
        comment3 = self.Comment(content=u'not cool dude', author=person3)
        self.session.add_all([person1, person2, person3])
        self.session.add_all([comment1, comment2, comment3])
        self.session.commit()
        # Search for any people who have comments that contain the word "cool".
        filters = [dict(name='comments', op='any',
                        val=dict(name='content', op='like', val='%cool%'))]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['1', '3'] == sorted(person['id'] for person in people)

    def test_has_in_to_one(self):
        """Test for filtering using the ``has`` operator with a sub-filter
        object on a to-one relationship.

        """
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        comment1 = self.Comment(content=u"that's cool!", author=person1)
        comment2 = self.Comment(content=u"i like turtles", author=person2)
        comment3 = self.Comment(content=u"not cool dude", author=person3)
        self.session.add_all([person1, person2, person3])
        self.session.add_all([comment1, comment2, comment3])
        self.session.commit()
        # Search for any comments whose author has ID equals to 1.
        filters = [dict(name='author', op='has',
                        val=dict(name='id', op='gt', val=1))]
        response = self.search('/api/comment', filters)
        document = loads(response.data)
        comments = document['data']
        assert ['2', '3'] == sorted(comment['id'] for comment in comments)

    def test_has_with_has(self):
        """Tests for nesting a ``has`` filter beneath another ``has`` filter.

        """
        for i in range(5):
            person = self.Person(id=i)
            article = self.Article(id=i)
            comment = self.Comment(id=i)
            article.author = person
            comment.author = person
            comment.article = article
            self.session.add_all([article, person, comment])
        self.session.commit()
        # Search for any comments whose articles have authors with id < 3.
        id_filter = dict(name='id', op='lt', val=3)
        author_filter = dict(name='author', op='has', val=id_filter)
        article_filter = dict(name='article', op='has', val=author_filter)
        filters = [article_filter]
        response = self.search('/api/comment', filters)
        document = loads(response.data)
        comments = document['data']
        assert ['0', '1', '2'] == sorted(comment['id'] for comment in comments)

    def test_any_with_any(self):
        """Tests for nesting an ``any`` filter beneath another ``any`` filter.

        """
        for i in range(5):
            person = self.Person(id=i)
            article = self.Article(id=i)
            comment = self.Comment(id=i)
            article.author = person
            comment.author = person
            comment.article = article
            self.session.add_all([article, person, comment])
        self.session.commit()
        # Search for any people whose articles have any comment with id < 3.
        id_filter = dict(name='id', op='lt', val=3)
        comments_filter = dict(name='comments', op='any', val=id_filter)
        articles_filter = dict(name='articles', op='any', val=comments_filter)
        filters = [articles_filter]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['0', '1', '2'] == sorted(person['id'] for person in people)

    def test_has_with_any(self):
        """Tests for nesting a ``has`` filter beneath an ``any`` filter."""
        for i in range(5):
            person = self.Person(id=i)
            article = self.Article(id=i)
            comment = self.Comment(id=i)
            article.author = person
            comment.author = person
            comment.article = article
            self.session.add_all([article, person, comment])
        self.session.commit()
        # Search for any articles with comments whose author has id < 3.
        id_filter = dict(name='id', op='lt', val=3)
        author_filter = dict(name='author', op='has', val=id_filter)
        comments_filter = dict(name='comments', op='any', val=author_filter)
        filters = [comments_filter]
        response = self.search('/api/article', filters)
        document = loads(response.data)
        articles = document['data']
        assert ['0', '1', '2'] == sorted(article['id'] for article in articles)

    def test_any_with_has(self):
        """Tests for nesting an ``any`` filter beneath a ``has`` filter."""
        for i in range(5):
            person = self.Person(id=i)
            article = self.Article(id=i)
            content = u'me' if i % 2 else u'you'
            comment = self.Comment(id=i, content=content)
            article.author = person
            comment.author = person
            self.session.add_all([article, person, comment])
        self.session.commit()
        # Search for any articles with an author who has made comments that
        # include the word "me".
        content_filter = dict(name='content', op='like', val='%me%')
        comment_filter = dict(name='comments', op='any', val=content_filter)
        author_filter = dict(name='author', op='has', val=comment_filter)
        filters = [author_filter]
        response = self.search('/api/article', filters)
        document = loads(response.data)
        articles = document['data']
        assert ['1', '3'] == sorted(article['id'] for article in articles)

    def test_comparing_fields(self):
        """Test for comparing the value of two fields in a filter object."""
        person1 = self.Person(id=1, age=1)
        person2 = self.Person(id=2, age=3)
        person3 = self.Person(id=3, age=3)
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='age', op='eq', field='id')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['1', '3'] == sorted(person['id'] for person in people)

    def test_date_yyyy_mm_dd(self):
        """Test for date parsing in filter objects with dates of the form
        ``1969-07-20``.

        """
        person1 = self.Person(id=1, birthday=date(1969, 7, 20))
        person2 = self.Person(id=2, birthday=date(1900, 1, 2))
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='birthday', op='eq', val='1969-07-20')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['1'] == sorted(person['id'] for person in people)

    def test_date_english(self):
        """Tests for date parsing in filter object with dates of the form ``2nd
        Jan 1900``.

        """
        person1 = self.Person(id=1, birthday=date(1969, 7, 20))
        person2 = self.Person(id=2, birthday=date(1900, 1, 2))
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='birthday', op='eq', val='2nd Jan 1900')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_times(self):
        """Test for time parsing in filter objects."""
        person1 = self.Person(id=1, bedtime=time(17))
        person2 = self.Person(id=2, bedtime=time(19))
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='bedtime', op='eq', val='19:00')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_datetimes(self):
        """Test for datetime parsing in filter objects."""
        person1 = self.Person(id=1, birth_datetime=datetime(1900, 1, 2))
        person2 = self.Person(id=2, birth_datetime=datetime(1969, 7, 20))
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='birth_datetime', op='eq', val='1969-07-20')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_datetime_to_date(self):
        """Tests that a filter object with a datetime value and a field with a
        ``Date`` type automatically converts the datetime to a date.

        """
        person1 = self.Person(id=1, birthday=date(1969, 7, 20))
        person2 = self.Person(id=2, birthday=date(1900, 1, 2))
        self.session.add_all([person1, person2])
        self.session.commit()
        datestring = '2nd Jan 1900 14:35'
        filters = [dict(name='birthday', op='eq', val=datestring)]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_datetime_to_time(self):
        """Test that a datetime gets truncated to a time if the model has a
        time field.

        """
        person1 = self.Person(id=1, bedtime=time(1, 2))
        person2 = self.Person(id=2, bedtime=time(14, 35))
        self.session.add_all([person1, person2])
        self.session.commit()
        datetimestring = datetime(1900, 1, 2, 14, 35).isoformat()
        filters = [dict(name='bedtime', op='eq', val=datetimestring)]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_bad_date(self):
        """Tests that an invalid date causes an error."""
        filters = [dict(name='birthday', op='eq', val='bogus')]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check error message here

    def test_bad_time(self):
        """Tests that an invalid time causes an error."""
        filters = [dict(name='bedtime', op='eq', val='bogus')]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check error message here

    def test_bad_datetime(self):
        """Tests that an invalid datetime causes an error."""
        filters = [dict(name='created_at', op='eq', val='bogus')]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check error message here

    def test_bad_name(self):
        """Tests that an invalid ``name`` element causes an error."""
        filters = [dict(name='bogus__field', op='eq', val='whatever')]
        response = self.search('/api/person', filters)
        check_sole_error(response, 400, ['No such field', 'bogus__field'])

    def test_search_boolean_formula(self):
        """Tests for Boolean formulas of filters in a search query."""
        person1 = self.Person(id=1, name=u'John', age=10)
        person2 = self.Person(id=2, name=u'Paul', age=20)
        person3 = self.Person(id=3, name=u'Luke', age=30)
        person4 = self.Person(id=4, name=u'Matthew', age=40)
        self.session.add_all([person1, person2, person3, person4])
        self.session.commit()
        # This searches for people whose name is John, or people older than age
        # 10 who have a "u" in their names. This should return three people:
        # John, Paul, and Luke.
        filters = [{'or': [{'and': [dict(name='name', op='like', val='%u%'),
                                    dict(name='age', op='ge', val=10)]},
                           dict(name='name', op='eq', val='John')]
                    }]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert len(people) == 3
        assert ['1', '2', '3'] == sorted(person['id'] for person in people)

    def test_dates_in_boolean_formulas(self):
        """Tests that dates are correctly handled in recursively defined
        boolean formula filters.

        For more information, see issue #423.

        """
        person1 = self.Person(id=1, birthday=date(1990, 1, 1))
        person2 = self.Person(id=2, birthday=date(1991, 1, 1))
        person3 = self.Person(id=3, birthday=date(1992, 1, 1))
        person4 = self.Person(id=4, birthday=date(1993, 1, 1))
        self.session.add_all([person1, person2, person3, person4])
        self.session.commit()
        filters = [{
            'and': [
                {
                    'name': 'birthday',
                    'op': '>',
                    'val': '1990-1-1'
                },
                {
                    'name': 'birthday',
                    'op': '<',
                    'val': '1993-1-1'
                }]
        }]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert len(people) == 2
        assert ['2', '3'] == sorted(person['id'] for person in people)

    @unittest.skip("I'm not certain in what situations an invalid value should cause a SQLAlchemy error")
    def test_invalid_value(self):
        """Tests for an error response on an invalid value in a filter object.

        """
        filters = [dict(name='age', op='>', val='should not be a string')]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check the error message here

    def test_invalid_field(self):
        """Tests for an error response on an invalid field name in a filter
        object.

        """
        filters = [dict(name='foo', op='>', val=2)]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check the error message here

    def test_invalid_operator(self):
        """Tests for an error response on an invalid operator in a filter
        object.

        """
        filters = [dict(name='age', op='bogus', val=2)]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check the error message here

    def test_missing_argument(self):
        """Tests that filter requests with a missing ``'val'`` causes an error
        response.

        """
        filters = [dict(name='name', op='==')]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check error message here

    def test_missing_fieldname(self):
        """Tests that filter requests with a missing ``'name'`` causes an error
        response.

        """
        filters = [dict(op='==', val='foo')]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check error message here

    def test_missing_operator(self):
        """Tests that filter requests with a missing ``'op'`` causes an error
        response.

        """
        filters = [dict(name='age', val=3)]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check error message here

    def test_to_many_relation(self):
        """Tests for filtering a to-many relation."""
        person = self.Person(id=1)
        articles = [self.Article(id=i) for i in range(5)]
        person.articles = articles
        self.session.add(person)
        self.session.add_all(articles)
        self.session.commit()
        filters = [dict(name='id', op='gt', val=2)]
        response = self.search('/api/person/1/articles', filters)
        document = loads(response.data)
        articles = document['data']
        assert ['3', '4'] == sorted(article['id'] for article in articles)


class TestOperators(SearchTestBase):

    def setUp(self):
        """Creates the database, the :class:`~flask.Flask` object, the
        :class:`~flask_restless.manager.APIManager` for that application,
        and creates the ReSTful API endpoints for the models used in the test
        methods.

        """
        super(TestOperators, self).setUp()

        class Person(self.Base):
            __tablename__ = 'person'
            id = Column(Integer, primary_key=True)
            name = Column(Unicode)
            # age = Column(Integer)
            # birthday = Column(Date)

        # class Comment(self.Base):
        #     __tablename__ = 'comment'
        #     id = Column(Integer, primary_key=True)
        #     content = Column(Unicode)
        #     author_id = Column(Integer, ForeignKey('person.id'))
        #     author = relationship('Person', backref=backref('comments'))

        self.Person = Person
        # self.Comment = Comment
        self.Base.metadata.create_all()
        self.manager.create_api(Person)
        # HACK Need to create APIs for these other models because otherwise
        # we're not able to create the link URLs to them.
        #
        # TODO Fix this by simply not creating links to related models for
        # which no API has been made.
        # self.manager.create_api(Comment)

    def test_equals(self):
        """Tests for the ``eq`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        self.session.add_all([person1, person2])
        self.session.commit()
        for op in '==', 'eq', 'equals', 'equal_to':
            filters = [dict(name='id', op=op, val=1)]
            response = self.search('/api/person', filters)
            document = loads(response.data)
            people = document['data']
            assert ['1'] == sorted(person['id'] for person in people)

    def test_not_equal(self):
        """Tests for the ``neq`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        self.session.add_all([person1, person2])
        self.session.commit()
        for op in '!=', 'ne', 'neq', 'not_equal_to', 'does_not_equal':
            filters = [dict(name='id', op=op, val=1)]
            response = self.search('/api/person', filters)
            document = loads(response.data)
            people = document['data']
            assert ['2'] == sorted(person['id'] for person in people)

    def test_greater_than(self):
        """Tests for the ``gt`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        self.session.add_all([person1, person2])
        self.session.commit()
        for op in '>', 'gt':
            filters = [dict(name='id', op=op, val=1)]
            response = self.search('/api/person', filters)
            document = loads(response.data)
            people = document['data']
            assert ['2'] == sorted(person['id'] for person in people)

    def test_less_than(self):
        """Tests for the ``lt`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        self.session.add_all([person1, person2])
        self.session.commit()
        for op in '<', 'lt':
            filters = [dict(name='id', op=op, val=2)]
            response = self.search('/api/person', filters)
            document = loads(response.data)
            people = document['data']
            assert ['1'] == sorted(person['id'] for person in people)

    def test_greater_than_or_equal(self):
        """Tests for the ``gte`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        for op in '>=', 'ge', 'gte', 'geq':
            filters = [dict(name='id', op=op, val=2)]
            response = self.search('/api/person', filters)
            document = loads(response.data)
            people = document['data']
            assert ['2', '3'] == sorted(person['id'] for person in people)

    def test_less_than_or_equal(self):
        """Tests for the ``lte`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        for op in '<=', 'le', 'lte', 'leq':
            filters = [dict(name='id', op=op, val=2)]
            response = self.search('/api/person', filters)
            document = loads(response.data)
            people = document['data']
            assert ['1', '2'] == sorted(person['id'] for person in people)

    def test_like(self):
        """Tests for the ``like`` operator."""
        person1 = self.Person(name=u'foo')
        person2 = self.Person(name=u'bar')
        person3 = self.Person(name=u'baz')
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='name', op='like', val='%ba%')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['bar', 'baz'] == sorted(person['attributes']['name']
                                        for person in people)

    def test_not_like(self):
        """Tests for the ``not_like`` operator."""
        person1 = self.Person(name=u'foo')
        person2 = self.Person(name=u'bar')
        person3 = self.Person(name=u'baz')
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='name', op='not_like', val='%fo%')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['bar', 'baz'] == sorted(person['attributes']['name']
                                        for person in people)

    def test_ilike(self):
        """Tests for the ``ilike`` operator."""
        person1 = self.Person(name=u'foo')
        person2 = self.Person(name=u'bar')
        person3 = self.Person(name=u'baz')
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='name', op='ilike', val='%BA%')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['bar', 'baz'] == sorted(person['attributes']['name']
                                        for person in people)

    def test_in(self):
        """Tests for the ``in`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='id', op='in', val=[1, 3])]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['1', '3'] == sorted(person['id'] for person in people)

    def test_not_in(self):
        """Tests for the ``not_in`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2)
        person3 = self.Person(id=3)
        self.session.add_all([person1, person2, person3])
        self.session.commit()
        filters = [dict(name='id', op='not_in', val=[1, 3])]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_is_null(self):
        """Tests for the ``is_null`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2, name=u'foo')
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='name', op='is_null')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['1'] == sorted(person['id'] for person in people)

    def test_is_not_null(self):
        """Tests for the ``is_not_null`` operator."""
        person1 = self.Person(id=1)
        person2 = self.Person(id=2, name=u'foo')
        self.session.add_all([person1, person2])
        self.session.commit()
        filters = [dict(name='name', op='is_not_null')]
        response = self.search('/api/person', filters)
        document = loads(response.data)
        people = document['data']
        assert ['2'] == sorted(person['id'] for person in people)

    def test_compare_equals_to_null(self):
        """Tests that an attempt to compare the value of a field to ``None``
        using the ``eq`` operator yields an error response, indicating that the
        user should use the ``is_null` operation instead.

        """
        filters = [dict(name='name', op='eq', val=None)]
        response = self.search('/api/person', filters)
        assert response.status_code == 400
        # TODO check the error message here.


class TestNetworkOperators(SearchTestBase):
    """Unit tests for the network address operators in PostgreSQL.

    For more information, see `Network Address Functions and Operators`_
    in the PostgreSQL documentation.

    .. _Network Address Functions and Operators: http://www.postgresql.org/docs/current/interactive/functions-net.html

    """

    def setUp(self):
        super(TestNetworkOperators, self).setUp()

        class Network(self.Base):
            __tablename__ = 'network'
            id = Column(Integer, primary_key=True)
            address = Column(INET)

        self.Network = Network
        self.Base.metadata.create_all()
        self.manager.create_api(Network)

    def tearDown(self):
        """Closes the database and removes the temporary directory in
        which it lives.

        """
        super(TestNetworkOperators, self).tearDown()
        self.database.stop()

    # We know this method will be called by `setUp()` in the superclass,
    # so we can set up the temporary database here.
    def database_uri(self):
        """Creates a PostgreSQL database and returns its connection URI."""
        #: The PostgreSQL database used by the test methods in this class.
        #:
        #: This attribute stores a
        #: :class:`~testing.postgresql.Postgresql` object, which must be
        #: stopped in the :meth:`.tearDown` method.
        self.database = PostgreSQL()

        return self.database.url()

    def test_is_not_equal(self):
        """Test for the ``<>`` ("is not equal") operator.

        For example:

        .. sourcecode:: postgresql

           inet '192.168.1.5' <> inet '192.168.1.4'

        """
        network1 = self.Network(id=1, address='192.168.1.5')
        network2 = self.Network(id=2, address='192.168.1.4')
        self.session.add_all([network1, network2])
        self.session.commit()
        filters = [dict(name='address', op='<>', val='192.168.1.4')]
        response = self.search('/api/network', filters)
        document = loads(response.data)
        networks = document['data']
        assert ['1'] == sorted(network['id'] for network in networks)

    def test_is_contained_by(self):
        """Test for the ``<<`` ("is contained by") operator.

        For example:

        .. sourcecode:: postgresql

           inet '192.168.1.5' << inet '192.168.1/24'

        """
        network1 = self.Network(id=1, address='192.168.1.5')
        network2 = self.Network(id=2, address='192.168.2.1')
        self.session.add_all([network1, network2])
        self.session.commit()
        filters = [dict(name='address', op='<<', val='192.168.1/24')]
        response = self.search('/api/network', filters)
        document = loads(response.data)
        networks = document['data']
        assert ['1'] == sorted(network['id'] for network in networks)

    def test_is_contained_by_or_equals(self):
        """Test for the ``<<=`` ("is contained by or equals") operator.

        For example:

        .. sourcecode:: postgresql

           inet '192.168.1/24' <<= inet '192.168.1/24'

        """
        network1 = self.Network(id=1, address='192.168.1/24')
        network2 = self.Network(id=2, address='192.168.1.5')
        network3 = self.Network(id=3, address='192.168.2.1')
        self.session.add_all([network1, network2, network3])
        self.session.commit()
        filters = [dict(name='address', op='<<=', val='192.168.1/24')]
        response = self.search('/api/network', filters)
        document = loads(response.data)
        networks = document['data']
        assert ['1', '2'] == sorted(network['id'] for network in networks)

    def test_contains(self):
        """Test for the ``>>`` ("contains") operator.

        For example:

        .. sourcecode:: postgresql

           inet '192.168.1/24' >> inet '192.168.1.5'

        """
        network1 = self.Network(id=1, address='192.168.1/24')
        network2 = self.Network(id=2, address='192.168.2/24')
        self.session.add_all([network1, network2])
        self.session.commit()
        filters = [dict(name='address', op='>>', val='192.168.1.5')]
        response = self.search('/api/network', filters)
        document = loads(response.data)
        networks = document['data']
        assert ['1'] == sorted(network['id'] for network in networks)

    def test_contains_or_equals(self):
        """Test for the ``>>=`` ("contains or equals") operator.

        For example:

        .. sourcecode:: postgresql

           inet '192.168.1/24' >>= inet '192.168.1/24'

        """
        network1 = self.Network(id=1, address='192.168.1/24')
        network2 = self.Network(id=2, address='192.168/16')
        network3 = self.Network(id=3, address='192.168.2/24')
        self.session.add_all([network1, network2, network3])
        self.session.commit()
        filters = [dict(name='address', op='>>=', val='192.168.1/24')]
        response = self.search('/api/network', filters)
        document = loads(response.data)
        networks = document['data']
        assert ['1', '2'] == sorted(network['id'] for network in networks)

    def test_contains_or_is_contained_by(self):
        """Test for the ``&&`` ("contains or is contained by") operator.

        .. warning::

           This operation is only available in PostgreSQL 9.4 or later.

        For example:

        .. sourcecode:: postgresql

           inet '192.168.1/24' && inet '192.168.1.80/28'

        """
        # network1 contains the queried subnet
        network1 = self.Network(id=1, address='192.168.1/24')
        # network2 is contained by the queried subnet
        network2 = self.Network(id=2, address='192.168.1.81/28')
        # network3 is neither
        network3 = self.Network(id=3, address='192.168.2.1')
        self.session.add_all([network1, network2, network3])
        self.session.commit()
        filters = [dict(name='address', op='&&', val='192.168.1.80/28')]
        response = self.search('/api/network', filters)
        document = loads(response.data)
        networks = document['data']
        assert ['1', '2'] == sorted(network['id'] for network in networks)


class TestAssociationProxy(SearchTestBase):
    """Test for filtering on association proxies."""

    def setUp(self):
        """Creates the database, the :class:`~flask.Flask` object, the
        :class:`~flask_restless.manager.APIManager` for that application,
        and creates the ReSTful API endpoints for the models used in the test
        methods.

        """
        super(TestAssociationProxy, self).setUp()

        class Article(self.Base):
            __tablename__ = 'article'
            id = Column(Integer, primary_key=True)
            tags = association_proxy('articletags', 'tag',
                                     creator=lambda tag: ArticleTag(tag=tag))

        class ArticleTag(self.Base):
            __tablename__ = 'articletag'
            article_id = Column(Integer, ForeignKey('article.id'),
                                primary_key=True)
            article = relationship(Article, backref=backref('articletags'))
            tag_id = Column(Integer, ForeignKey('tag.id'), primary_key=True)
            tag = relationship('Tag')
            # TODO this dummy column is required to create an API for this
            # object.
            id = Column(Integer)

        class Tag(self.Base):
            __tablename__ = 'tag'
            id = Column(Integer, primary_key=True)
            name = Column(Unicode)

        self.Article = Article
        self.Tag = Tag
        self.Base.metadata.create_all()
        self.manager.create_api(Article)
        # HACK Need to create APIs for these other models because otherwise
        # we're not able to create the link URLs to them.
        #
        # TODO Fix this by simply not creating links to related models for
        # which no API has been made.
        self.manager.create_api(ArticleTag)
        self.manager.create_api(Tag)

    def test_any(self):
        """Tests for filtering on a many-to-many relationship via an
        association proxy backed by an association object.

        """
        article1 = self.Article(id=1)
        article2 = self.Article(id=2)
        article3 = self.Article(id=3)
        tag1 = self.Tag(name=u'foo')
        tag2 = self.Tag(name=u'bar')
        tag3 = self.Tag(name=u'baz')
        article1.tags = [tag1, tag2]
        article2.tags = [tag2, tag3]
        article3.tags = [tag3, tag1]
        self.session.add_all([article1, article2, article3])
        self.session.add_all([tag1, tag2, tag3])
        self.session.commit()
        filters = [dict(name='tags', op='any',
                        val=dict(name='name', op='eq', val='bar'))]
        response = self.search('/api/article', filters)
        document = loads(response.data)
        articles = document['data']
        assert ['1', '2'] == sorted(article['id'] for article in articles)



import cProfile
import pstats
import time
import unittest

from flask import Flask
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import Unicode
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker

from flask_restless import APIManager

Base = declarative_base()


class Person(Base):
    __tablename__ = 'person'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode)


class Article(Base):
    __tablename__ = 'article'

    id = Column(Integer, primary_key=True)
    title = Column(String(16))
    author_id = Column(Integer, ForeignKey('person.id'))
    author = relationship('Person', backref=backref('articles'))


@unittest.skip("Slow test, for manual run only")
class TestPerformance(unittest.TestCase):

    def setUp(self):
        app = Flask(__name__)
        self.engine = create_engine('sqlite://', echo=False)
        session = scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=self.engine))

        Base.metadata.create_all(bind=self.engine)
        session.bulk_save_objects([Person(id=i, name=f'Person {i}') for i in range(10000)])
        session.bulk_save_objects([Article(id=i, title=f'Title {i}', author_id=1) for i in range(10000)])

        self.test_client = app.test_client()

        api_manager = APIManager(app=app, session=session, url_prefix='/api')
        api_manager.create_api(Person, collection_name='people', page_size=0)
        api_manager.create_api(Article, collection_name='articles', page_size=0)
        self.profile = cProfile.Profile()

    def tearDown(self):
        Base.metadata.drop_all(self.engine)

    def test_loading_collection(self):
        self.profile.enable()
        start_time = time.time()
        response = self.test_client.get('/api/articles')
        self.profile.disable()
        processing_time = time.time() - start_time
        print('Fetch 10000 articles time:', processing_time)
        stats = pstats.Stats(self.profile)
        stats.sort_stats('time', 'cumulative').print_stats(20)
        assert response.status_code == 200
